!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|
!                                                                              |
!              ||===\\                                                         | 
!              ||    \\                                                        |
!              ||     ||   //==\\   ||  ||   //==||  ||/==\\                   |
!              ||     ||  ||    ||  ||  ||  ||   ||  ||    ||                  |
!              ||    //   ||    ||  ||  ||  ||   ||  ||                        |
!              ||===//     \\==//    \\==\\  \\==\\  ||                        |
!                                                                              |
!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|
!                                                                              |
!              UPDATE_TRCLOUD              Jan. 2015                           |
!                                                                              |
!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|

subroutine update_trcloud (trcl,osolve,params,istep,dt,current_time) 

!------------------------------------------------------------------------------|
!(((((((((((((((( Purpose of the routine  ))))))))))))))))))))))))))))))))))))))
!------------------------------------------------------------------------------|
! this file contains the operations done to 'groom' the tracking cloud - check 
! the position of particles, save surfaced ones, discard others and inject new 
! ones.
! In the first run, the cloud is created.

!------------------------------------------------------------------------------|
!((((((((((((((((  declaration of the subroutine arguments  ))))))))))))))))))))
!------------------------------------------------------------------------------|
      
use definitions
!use mpi

implicit none

include 'mpif.h'

type (tracking_cloud) trcl
type (octreesolve) osolve
type (parameters) params
integer istep
double precision dt,current_time

!------------------------------------------------------------------------------|
!(((((((((((((((( declaration of the subroutine internal variables )))))))))))))
!------------------------------------------------------------------------------|

integer iproc,nproc,ierr,err,iter
integer i,j,k,trgridshape,gridstep,tracksi,ie
integer trstep,ntrpx,ntrpy,ntrpz
integer nsurfd,nreset,leaf,level,locc
integer,dimension(:),allocatable::trdensloss
integer,dimension(:,:),allocatable::trackswap
double precision u,v,w,rat
double precision x0,y0,z0,xi,yi,zi,x,y,z,dxyz
double precision,dimension(3) :: pos
double precision,dimension(:),allocatable::xbuf,ybuf,zbuf

double precision eps,dx,dy,dz
double precision,dimension(:),allocatable::lsf0,lsf0_old,temp,press,lsf

character(4) cistep
character(10) outdir
!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|

call mpi_comm_size (mpi_comm_world,nproc,ierr)
call mpi_comm_rank (mpi_comm_world,iproc,ierr)

!------------------------------------------------------------------------------|
! Creating Tracking Cloud at first timestep
!------------------------------------------------------------------------------|
if (trcl%np == 0) then
    !if not loading from restart file, create tracking particles according to 
    !specifications in input file
    
    ntrpx = params%ntrpx
    ntrpy = params%ntrpy
    ntrpz = params%ntrpz

    injection: if (params%gridtrack) then
    !Regularly gridded tracking cloud
        trgridshape = mod(params%ntrpx,2)+mod(params%ntrpy,2)+mod(params%ntrpy,2)
        reggrid: if (trgridshape==0) then
        !simple cubic grid
            dx = (params%trboxxmax-params%trboxxmin)/ntrpx
            dy = (params%trboxymax-params%trboxymin)/ntrpy
            dz = (params%trboxzmax-params%trboxzmin)/ntrpz

            !create tracking particles
            gridstep = 1
            do i = 1,ntrpx
                do j = 1,ntrpy
                    do k = 1,ntrpz
                        trcl%x(gridstep) = params%trboxxmin+(i-0.5d0)*dx
                        trcl%y(gridstep) = params%trboxymin+(j-0.5d0)*dy
                        trcl%z(gridstep) = params%trboxzmin+(k-0.5d0)*dz
                        gridstep = gridstep+1
                    end do
                end do
            end do

        else reggrid
        !trgridshape=3, bcc-type gric
            dx = 2.d0*(params%trboxxmax-params%trboxxmin)/(ntrpx+1)
            dy = 2.d0*(params%trboxymax-params%trboxymin)/(ntrpy+1)
            dz = 2.d0*(params%trboxzmax-params%trboxzmin)/(ntrpz+1)

            !create tracking particles
            gridstep = 1
            do i = 1,(ntrpx+1)/2
                do j = 1,(ntrpy+1)/2
                    do k = 1,(ntrpz+1)/2
                        trcl%x(gridstep) = params%trboxxmin+(i-0.5d0)*dx
                        trcl%y(gridstep) = params%trboxymin+(j-0.5d0)*dy
                        trcl%z(gridstep) = params%trboxzmin+(k-0.5d0)*dz
                        gridstep = gridstep+1
                    end do
                end do
            end do
            do i = 1,(ntrpx-1)/2
                do j = 1,(ntrpy-1)/2
                    do k = 1,(ntrpz-1)/2
                        trcl%x(gridstep) = params%trboxxmin+i*dx
                        trcl%y(gridstep) = params%trboxymin+j*dy
                        trcl%z(gridstep) = params%trboxzmin+k*dz
                        gridstep = gridstep+1
                    end do
                end do
            end do
        end if reggrid

        !for testing puproses, remove afterwards
        if (gridstep-1 == params%ntrp) then
            trcl%np = params%ntrp
        else
            call stop_run ('error in tracking cloud grid injection')
        end if
        
    else injection
    !Randomly distributed tracking cloud.
    !CAUTION: Random would be different for each processor, so position has to be
    !calculated centrally and then distributed.
        write(*,*) ' using random particle creation'
        if (iproc==0) then
        !calculating particle positions on master process
            call random_number (trcl%x)
            call random_number (trcl%y)
            call random_number (trcl%z)
            do i = 1,params%ntrp
                trcl%x(i) = params%trboxxmin+(params%trboxxmax-params%trboxxmin)*trcl%x(i)
                trcl%y(i) = params%trboxymin+(params%trboxymax-params%trboxymin)*trcl%y(i)
                trcl%z(i) = params%trboxzmin+(params%trboxzmax-params%trboxzmin)*trcl%z(i)
            end do
        end if  

        !Distributing particle position to all processes
        trcl%np = params%ntrp
        call mpi_bcast(trcl%x,trcl%np,mpi_double_precision,0,mpi_comm_world,ierr)
        call mpi_bcast(trcl%y,trcl%np,mpi_double_precision,0,mpi_comm_world,ierr)
        call mpi_bcast(trcl%z,trcl%np,mpi_double_precision,0,mpi_comm_world,ierr)
    end if injection

    !save initial particle positions
    trcl%x0 = trcl%x
    trcl%y0 = trcl%y
    trcl%z0 = trcl%z

    !create reference array for particle density
    trcl%ne_init = osolve%nleaves
    allocate (trcl%trdens_init(trcl%ne_init),stat=err) ; if (err.ne.0) call stop_run ('error in alloc trdens_init in update_trcloud$')

    trcl%trdens_init=0
    do i=1,trcl%np
       call octree_find_leaf (osolve%octree,osolve%noctree,trcl%x0(i),trcl%y0(i),trcl%z0(i), &
                              leaf,level,locc,x0,y0,z0,dxyz)
       trcl%trdens_init(leaf)=trcl%trdens_init(leaf)+1
    enddo
end if


!------------------------------------------------------------------------------|
! Interpolate values from osolve onto tracking particles                       |
! -- taken form update_cloud_fields.f90                                        |
!------------------------------------------------------------------------------|

!Get upper surface level set function
allocate (lsf(osolve%nnode),stat=err) ; if (err.ne.0) call stop_run ('Error alloc lsf in update_cloud_fields$')
lsf=0.d0
if (osolve%nlsf.ne.0) lsf=osolve%lsf(1:osolve%nnode,1)

!allocate temporary arrays
allocate (lsf0(trcl%np),stat=err) ; if (err.ne.0) call stop_run ('Error alloc lsf0 in update_trcloud$')
!allocate (lsf0_old(trcl%np),stat=err) ; if (err.ne.0) call stop_run ('Error alloc lsf0_old in update_trcloud$')
allocate (temp(trcl%np),stat=err) ; if (err.ne.0) call stop_run ('Error alloc temp in update_trcloud$')
allocate (press(trcl%np),stat=err) ; if (err.ne.0) call stop_run ('Error alloc press in update_trcloud$')

!lsf0_old=lsf0
!Set to 0 here because allreduce sums over all partly calcualted arrays after calculation
lsf0=0.d0
temp=0.d0
press=0.d0

!interpolate values
do k=1+iproc,trcl%np,nproc
   call octree_interpolate_three (3,osolve%octree,osolve%noctree,osolve%icon, &
                                 osolve%nleaves,osolve%nnode, &
                                 trcl%x(k),trcl%y(k),trcl%z(k), &
                                 lsf,lsf0(k), &
                                 osolve%temp,temp(k), &
                                 osolve%spressure,press(k)) !Using smoothed pressure - okay to do so?
enddo

call mpi_allreduce (lsf0, trcl%lsf0, trcl%np, mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
call mpi_allreduce (temp, trcl%temp, trcl%np, mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
call mpi_allreduce (press,trcl%press,trcl%np, mpi_double_precision,mpi_sum,mpi_comm_world,ierr)

deallocate (lsf)
deallocate (lsf0)
!deallocate (lsf0_old)
deallocate (temp)
deallocate (press)

!------------------------------------------------------------------------------|
! Looking for surfaced particles (lsf0>0) and saving their track extent        |
!------------------------------------------------------------------------------|

!find number of surfaced particles
nsurfd = 0
do i = 1,trcl%np
    if (trcl%lsf0(i) > 0) nsurfd = nsurfd+1
end do

savetracks: if (nsurfd>0) then
    !increase tracks array for newly surfaced particles
    tracksize: if (trcl%ntracks > 0) then
        !allocate swap and cop y tracks
        allocate (trackswap(trcl%ntracks,3), stat=err) ; if (err.ne.0) call stop_run ('Error alloc trackswap in update_trcloud$')
        trackswap = trcl%tracks

        !delete and reallocate tracks of increased size
        deallocate (trcl%tracks)
        allocate (trcl%tracks(trcl%ntracks+nsurfd,3), stat=err) ; if (err.ne.0) call stop_run ('Error alloc trcl%tracks in define_trcloud$')

        !copy to new tracks and deallocate swap
        trcl%tracks(1:trcl%ntracks,:) = trackswap
        deallocate (trackswap)
    else tracksize
        !if no tracks ha ve been saved yet, trcl%ntracks should be 0, no swap is needed
        deallocate (trcl%tracks)
        allocate (trcl%tracks(nsurfd,3), stat=err) ; if (err.ne.0) call stop_run ('Error alloc trcl%tracks in define_trcloud$')
    end if tracksize

    !write new track information, reset is done in move section
    tracksi = trcl%ntracks+1
    do i = 1,trcl%np
        if (trcl%lsf0(i) > 0) then 
            trcl%tracks(tracksi,:)  = (/i,trcl%trackstart(i),istep/)
            tracksi = tracksi+1
        end if
    end do
    !update number of saved tracks
    trcl%ntracks = trcl%ntracks+nsurfd
    
end if savetracks
!------------------------------------------------------------------------------|
! Write tracking information to disk                                           |
! All position, pressure and temperature values are saved as reals.
! Additionally, tracks that surfaced in this timestep are saved as well.
!------------------------------------------------------------------------------|

if (iproc.eq.0) then
    call int_to_char (cistep,4,istep)
    open (20,file='TROUT/tracking_t'//cistep//'.bin',status='new',form='unformatted')

    !information on array size and time step 
    write (20) trcl%np,nsurfd,istep,current_time
 
    !scaling information
    write (20) params%lref,params%veloref,params%rhoref
 
    !write value arrays
    write (20) real(trcl%x)
    write (20) real(trcl%y)
    write (20) real(trcl%z*params%vex)
    write (20) real(trcl%temp)
    write (20) real(trcl%press)
 
    !write tracks of surfaced particles
    do i = 1,nsurfd
        write (20) trcl%tracks(trcl%ntracks-nsurfd+i,:)
    end do

    close (20)
end if


!------------------------------------------------------------------------------|
! Move tracking particles                                                      |
! taken from move_cloud.f90                                                    |
!------------------------------------------------------------------------------|

allocate (xbuf(trcl%np),stat=err) ; if (err.ne.0) call stop_run ('Error alloc xbuf in update_trcloud$')
allocate (ybuf(trcl%np),stat=err) ; if (err.ne.0) call stop_run ('Error alloc ybuf in update_trcloud$')
allocate (zbuf(trcl%np),stat=err) ; if (err.ne.0) call stop_run ('Error alloc zbuf in update_trcloud$')
 
xbuf=0.d0
ybuf=0.d0
zbuf=0.d0
rat=1.d0-1.d-10

do i=1+iproc,trcl%np,nproc
    if (trcl%lsf0(i) <= 0.d0) then
    !For particles not surfaced, move according to interpolation of nodal velocities
        u=0.d0
        v=0.d0
        w=0.d0
        x0=trcl%x(i)*rat
        y0=trcl%y(i)*rat
        z0=trcl%z(i)*rat
        x=x0
        y=y0
        z=z0
        !midpoint algorithm 
        do iter=1,params%niter_move
           xi=(x0+x)/2.d0
           yi=(y0+y)/2.d0
           zi=(z0+z)/2.d0
           call octree_interpolate_three(3,osolve%octree,osolve%noctree,osolve%icon,osolve%nleaves,osolve%nnode,xi,yi,zi,osolve%u,u,osolve%v,v,osolve%w/params%vex,w)
           !Move only if particle is within domain - really necessary?
           if (trcl%x(i)*(trcl%x(i)-1.d0).lt.0.d0) x=x0+dt*u
           if (trcl%y(i)*(trcl%y(i)-1.d0).lt.0.d0) y=y0+dt*v
           if (trcl%z(i)*(trcl%z(i)-1.d0).lt.0.d0) z=z0+dt*w
        enddo
        xbuf(i)=x
        ybuf(i)=y
        zbuf(i)=z
    else
    !if surfaced, do not move
        !move sufaced particles exactly to surface?
        xbuf(i)=trcl%x(i)
        ybuf(i)=trcl%y(i)
        zbuf(i)=trcl%z(i)
    end if
enddo

trcl%x=0.d0
trcl%y=0.d0
trcl%z=0.d0
call mpi_allreduce (xbuf,trcl%x,trcl%np,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
call mpi_allreduce (ybuf,trcl%y,trcl%np,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
call mpi_allreduce (zbuf,trcl%z,trcl%np,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)

deallocate (xbuf)
deallocate (ybuf)
deallocate (zbuf)

!------------------------------------------------------------------------------|
! Reset surfaced and exited particles                                          |
!------------------------------------------------------------------------------|

! Check whether density referene array is still valid. If not, recalculate from 
! initial particle positions.

if (trcl%ne_init .ne. osolve%nleaves) then
    if (iproc == 0) write (*,*) 'Start Calulation reference array'
    !get rid of old array
    deallocate (trcl%trdens_init)

    !create new reference array for particle density
    trcl%ne_init = osolve%nleaves
    allocate (trcl%trdens_init(trcl%ne_init),stat=err) ; if (err.ne.0) call stop_run ('error in alloc trdens_init in update_trcloud$')

    trcl%trdens_init=0
    do i=1,trcl%np
       call octree_find_leaf (osolve%octree,osolve%noctree,trcl%x0(i),trcl%y0(i),trcl%z0(i), &
                              leaf,level,locc,x0,y0,z0,dxyz)
       trcl%trdens_init(leaf)=trcl%trdens_init(leaf)+1
    enddo
end if

! Random injection has to be done centrally, so the new coordinates are
! calculated on root process and then boradcasted (alternative: initialize 
! random number generator the same way for all processes)
reset: if (iproc==0) then
    !calculate the decrease in each element
    allocate (trdensloss(trcl%ne_init),stat=err) ; if (err.ne.0) call stop_run ('error in alloc trdensloss in update_trcloud$')

    trdensloss=0
    do i=1,trcl%np
       call octree_find_leaf (osolve%octree,osolve%noctree,trcl%x(i),trcl%y(i),trcl%z(i), &
                              leaf,level,locc,x0,y0,z0,dxyz)
       trdensloss(leaf)=trdensloss(leaf)+1
    enddo
    trdensloss = trcl%trdens_init-trdensloss

    nreset=0
    do i=1,trcl%np
        if (trcl%x(i)>1.d0 .OR. trcl%x(i)<0.d0 .OR. &
            trcl%y(i)>1.d0 .OR. trcl%y(i)<0.d0 .OR. &
            trcl%lsf0(i)>0.d0 .OR. trcl%z(i)<0.d0) then

            !find leaf with hightes loss of particles
            ie = maxloc(trdensloss,DIM=1)
            !adjust loss number in leaf
            trdensloss(ie) = trdensloss(ie)-1

            !inject randomly in this leaf
            call random_number(pos)
            pos(1)=osolve%x(osolve%icon(1,ie))+pos(1)*(osolve%x(osolve%icon(2,ie))-osolve%x(osolve%icon(1,ie)))
            pos(2)=osolve%y(osolve%icon(1,ie))+pos(2)*(osolve%y(osolve%icon(3,ie))-osolve%y(osolve%icon(1,ie)))
            pos(3)=osolve%z(osolve%icon(1,ie))+pos(3)*(osolve%z(osolve%icon(5,ie))-osolve%z(osolve%icon(1,ie)))

            trcl%x(i) = pos(1)
            trcl%y(i) = pos(2)
            trcl%z(i) = pos(3)

            !reset trackstart
            trcl%trackstart(i)= istep+1

            !send particle position to other processes
            call mpi_bcast(i,1,mpi_integer,0,mpi_comm_world,ierr)
            call mpi_bcast(pos,3,mpi_double_precision,0,mpi_comm_world,ierr)

            nreset = nreset+1
        end if
    end do 

    !send end signal to other processors
    call mpi_bcast(0,1,mpi_integer,0,mpi_comm_world,ierr)
    
    write (*,'(a,i3,a,i3,a)') 'update_trcloud: ',nsurfd,'particles surfaced and in total ',nreset,'particles reset'
    deallocate (trdensloss)

else reset
    !other proceses: wait for particle ID and new position, exit loop if end signal is recieved
    do
        !recieve ID or end signal
        call mpi_bcast(i,1,mpi_integer,0,mpi_comm_world,ierr)
        if (i==0) exit
        
        !recieve new position
        call mpi_bcast(pos,3,mpi_double_precision,0,mpi_comm_world,ierr)
        !apply new position
        trcl%x(i) = pos(1)
        trcl%y(i) = pos(2)
        trcl%z(i) = pos(3)
        !reset trackstart
        trcl%trackstart(i)= istep+1
    end do
end if reset

return
end
!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|
